from collections.abc import MutableMapping, MutableSequence
from typing import TYPE_CHECKING, Any, Callable, Iterator, Mapping, Sequence, TypeAlias

column_name: TypeAlias = str
column_value: TypeAlias = Any
